import datetime
import math
from string import *

from thunno2.codepage import *
from thunno2.version import *

"""
Constants in Thunno start with a `k` and then have one other character after that.
This is the dictionary where all the constants are defined
"""

CONSTANTS = {
    'A': ascii_uppercase,
    'B': 'Buzz',
    'C': CODEPAGE,
    'D': '0123456789',
    'E': math.e,
    'F': 'Fizz',
    'G': (1 + 5 ** 0.5) / 2,
    'H': 'Hello, World!',
    'I': datetime.datetime.now().microsecond,
    'J': datetime.datetime.now().second,
    'K': datetime.datetime.now().minute,
    'L': datetime.datetime.now().hour,
    'M': datetime.datetime.now().day,
    'N': datetime.datetime.now().month,
    'O': datetime.datetime.now().year,
    'P': math.pi,
    'Q': str(datetime.datetime.now()).split('.')[0],
    'R': '()[]{}',
    'S': '([{',
    'T': ')]}',
    'U': '([{<',
    'V': ')]}>',
    'W': 'AEIOU',
    'X': 'BCDFGHJKLMNPQRSTVWXYZ',
    'Y': 'AEIOUY',
    'Z': 'BCDFGHJKLMNPQRSTVWXZ',
    'a': ascii_lowercase + ascii_uppercase,
    'b': 'buzz',
    'c': ''.join('%c' * 95 % (*range(32, 127),)),
    'd': '01',
    'e': '01234567',
    'f': '0123456789ABCDEF',
    'g': ['Fizz', 'Buzz'],
    'h': 'Hello World',
    'i': 16,
    'j': 32,
    'k': 64,
    'l': 128,
    'm': 256,
    'n': 512,
    'o': 1024,
    'p': 2048,
    'q': 4096,
    'r': 8192,
    's': 16384,
    't': 32768,
    'u': 65536,
    'v': ["qwertyuiop", "asdfghjkl", "zxcvbnm"],
    'w': 'aeiou',
    'x': 'bcdfghjklmnpqrstvwxyz',
    'y': 'aeiouy',
    'z': 'bcdfghjklmnpqrstvwxz',
    '1': 10 ** 3,
    '2': 10 ** 4,
    '3': 10 ** 5,
    '4': 10 ** 6,
    '5': 10 ** 7,
    '6': 10 ** 8,
    '7': 10 ** 9,
    '8': 10 ** 10,
    '9': 10 ** 11,
    '0': 10 ** 12,
    '!': THUNNO_VERSION,
    '"': [0, 0],
    '#': [0, 1],
    '$': [0, -1],
    '%': [1, 0],
    '&': [1, 1],
    "'": [1, -1],
    '(': [-1, 0],
    ')': [-1, 1],
    '*': [-1, -1],
    '+': [3, 5],
    'Ȧ': ascii_lowercase + ascii_uppercase + '0123456789',
    'Ḃ': '0123456789' + ascii_uppercase + ascii_lowercase,
    'Ċ': ascii_lowercase + ascii_uppercase + '0123456789_',
    'Ḋ': ascii_uppercase[::-1],
    'Ė': ascii_lowercase[::-1] + ascii_uppercase[::-1],
    'Ḟ': ascii_lowercase[::-1] + ascii_uppercase[::-1] + '9876543210_',
    'Ġ': ascii_uppercase + ascii_lowercase,
    'Ḣ': ascii_uppercase[::-1] + ascii_lowercase[::-1],
    'İ': ascii_lowercase[::-1] + ascii_uppercase[::-1] + '9876543210',
    'Ŀ': ascii_uppercase + ascii_lowercase + '0123456789',
    'Ṁ': ascii_uppercase[::-1] + ascii_lowercase[::-1] + '9876543210',
    'Ṅ': '9876543210',
    'Ȯ': '\n' + ''.join('%c' * 95 % (*range(32, 127),)),
    'Ṗ': punctuation,
    'Ṙ': 'http://',
    'Ṡ': 'http://www.',
    'Ṫ': 'https://www.',
    'Ẇ': whitespace,
    'Ẋ': ['()', '[]', '{}', '<>'],
    'Ẏ': '([{<>}])',
    'Ż': '[]<>-+.,'
}



from thunno2 import lexer, version, flags
import sys


def from_terminal():
    print('Thunno', version.THUNNO_VERSION, 'interpreter\n')
    print('\nFlags:')
    flags_list = input()
    code = ''
    print('\nHeader:')
    inp = input()
    while inp:
        code += inp + '\n'
        inp = input()
    print('\nCode:')
    inp = input()
    while inp:
        code += inp + '\n'
        inp = input()
    print('\nFooter:')
    inp = input()
    while inp:
        code += inp + '\n'
        inp = input()
    inputs = ''
    print('\nInput:')
    inp = input()
    while inp:
        inputs += inp + '\n'
        inp = input()
    _, tokenised = lexer.tokenise(code)
    print('\nOutput:')
    flags.run(flags_list, tokenised, inputs)


def from_cmdline():
    args = sys.argv[1:]
    if not args:
        from_terminal()
        return None
    filename = args[0]
    flags_list = args[1:]
    sys.stderr.write('Thunno, v' + version.THUNNO_VERSION + '\n')
    try:
        with open(filename) as f:
            _, tokenised = lexer.tokenise(f.read())
            flags.run(''.join(flags_list), tokenised, sys.stdin.read())
    except Exception as E:
        sys.stderr.write('An error occurred: ' + repr(E))



# This codepage is exactly the same as the Jelly codepage
# (https://github.com/DennisMitchell/jellylanguage/blob/master/jelly/interpreter.py#L7)

"""The 256-character Single Byte Character Set for encoding Thunno 2 programs (https://en.wikipedia.org/wiki/SBCS)"""

CODEPAGE = r'''¡¢£¤¥¦©¬®µ½¿€ÆÇÐÑ×ØŒÞßæçðıȷñ÷øœþ !"#$%&'()*+,-./0123456789:;<=>?''' \
           r'''@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~¶''' \
           r'''°¹²³⁴⁵⁶⁷⁸⁹⁺⁻⁼⁽⁾ƁƇƊƑƓƘⱮƝƤƬƲȤɓƈɗƒɠɦƙɱɲƥʠɼʂƭʋȥẠḄḌẸḤỊḲḶṂṆỌṚṢṬỤṾẈỴẒȦḂ''' \
           r'''ĊḊĖḞĠḢİĿṀṄȮṖṘṠṪẆẊẎŻạḅḍẹḥịḳḷṃṇọṛṣṭ§Äẉỵẓȧḃċḋėḟġḣŀṁṅȯṗṙṡṫẇẋẏż«»‘’“”'''

assert len(CODEPAGE) == 256


def codepage_index(*chars):
    for char in chars:
        if char in CODEPAGE:
            yield CODEPAGE.index(char)
        else:
            yield -1


def utf8_to_thunno2(string):
    ret = ''
    for char in string:
        if ord(char) < 256:
            ret += CODEPAGE[ord(char)]
        else:
            ret += char
    return ret


def thunno2_to_utf8(string):
    ret = ''
    for char in string:
        if char in CODEPAGE:
            ret += chr(CODEPAGE.index(char))
        else:
            ret += char
    return ret



from thunno2 import interpreter, commands, helpers


def process_input_flags(flags, inputs):

    if 'w' in flags:
        commands.ctx.warnings = True

    if 'W' in flags:
        inputs = [inputs]
    else:
        inputs = inputs.splitlines()

    if 'E' not in flags:
        new_input = []
        for inp in inputs:
            try:
                x = eval(inp)
                if type(x) in (int, float, str, list):
                    new_input.append(x)
                elif isinstance(x, bool):
                    new_input.append(int(x))
                elif isinstance(x, (set, tuple)):
                    new_input.append(list(x))
                elif isinstance(x, dict):
                    new_input.append(list(map(list, x.items())))
                else:
                    new_input.append(inp)
            except:
                new_input.append(inp)
        inputs = new_input[:]

    for flag in flags:
        if flag == 'Z':
            commands.ctx.stack.push(0)
        elif flag == 'T':
            commands.ctx.stack.push(10)
        elif flag == 'H':
            commands.ctx.stack.push(100)

    if 'B' in flags:
        new_input = []
        for inp in inputs:
            if isinstance(inp, str):
                new_input.append([ord(c) for c in inp])
            else:
                new_input.append(inp)
        inputs = new_input[:]

    if '+' in flags:
        new_input = []
        for inp in inputs:
            if isinstance(inp, (int, float)):
                new_input.append(inp + flags.count('+'))
            else:
                new_input.append(inp)
        inputs = new_input[:]

    if '-' in flags:
        new_input = []
        for inp in inputs:
            if isinstance(inp, (int, float)):
                new_input.append(inp - flags.count('-'))
            else:
                new_input.append(inp)
        inputs = new_input[:]

    return inputs


def process_output_flags(flags):

    for flag in flags:

        if 'J' == flag:
            commands.ctx.stack.push(helpers.empty_join(next(commands.ctx.stack.rmv(1))))

        if 'j' == flag:
            commands.ctx.stack.push(helpers.empty_join(list(commands.ctx.stack)))

        if 'N' == flag:
            commands.ctx.stack.push(helpers.newline_join(next(commands.ctx.stack.rmv(1))))

        if 'n' == flag:
            commands.ctx.stack.push(helpers.newline_join(list(commands.ctx.stack)))

        if 'Ṡ' == flag:
            commands.ctx.stack.push(helpers.space_join(next(commands.ctx.stack.rmv(1))))

        if 'ṡ' == flag:
            commands.ctx.stack.push(helpers.space_join(list(commands.ctx.stack)))

        if 'S' == flag:
            commands.ctx.stack.push(commands.commands['S']()[0])

        if 's' == flag:
            commands.ctx.stack.push(helpers.it_sum(commands.ctx.stack))

        if 'L' == flag:
            commands.ctx.stack.push(commands.commands['l']()[0])

        if 'l' == flag:
            commands.ctx.stack.push(len(commands.ctx.stack))

        if 'h' == flag:
            commands.ctx.stack.push(commands.commands['h']()[0])

        if 't' == flag:
            commands.ctx.stack.push(commands.commands['t']()[0])

        if 'B' == flag:
            x = (commands.ctx.stack + commands.ctx.other_il + [0])[0]
            if isinstance(x, list):
                r = ''
                for i in x:
                    try:
                        r += chr(int(i))
                    except:
                        pass
                commands.ctx.stack.push(r)

    if (commands.ctx.implicit_print or ('O' in flags)) and not ('o' in flags):
        print(next(commands.ctx.stack.rmv(1)))


def run(flags, code, inputs):

    if 'V' in flags:
        new_flags = ''.join(f for f in flags if f != 'V')
        for line in inputs.splitlines():
            try:
                x = eval(line)
                if isinstance(x, tuple):
                    new_inputs = process_input_flags(new_flags, '\n'.join(map(repr, x)))
                else:
                    new_inputs = [x]
            except:
                new_inputs = [line]
            commands.ctx.og_input_list = new_inputs.copy()
            commands.ctx.other_il = new_inputs.copy()
            print(line, '--> ', end='')
            interpreter.run(code, n=0, iteration_index=0)
            process_output_flags(new_flags)
        return None

    inputs = process_input_flags(flags, inputs)
    commands.ctx.og_input_list = inputs.copy()
    commands.ctx.other_il = inputs.copy()
    interpreter.run(code, n=0, iteration_index=0)
    process_output_flags(flags)
